==> lx_anonymizer-0.8.2.2.tar.gz: lx_anonymizer-0.8.2.2/lx_anonymizer/ocr_tesserocr.py <==
"""
Optimized OCR module using Tesserocr for significantly improved performance.

This module replaces pytesseract usage with Tesserocr for:
- 10-50x faster text recognition in video processing
- Pre-loaded Tesseract model (memoization)
- Direct C++ API access without CLI overhead
- Better memory management for batch processing

Usage:
    from lx_anonymizer.ocr_tesserocr import tesseract_on_boxes_fast, TesseOCROptimized
    
    # Drop-in replacement for existing tesseract_on_boxes function
    results, confidences = tesseract_on_boxes_fast(image_path, boxes)
    
    # Or use the class directly for more control
    ocr_processor = TesseOCROptimized()
    results, confidences = ocr_processor.process_image_boxes(image_path, boxes)
"""

import logging
import time
from typing import List, Tuple, Union
import tesserocr
from PIL import Image
import threading

logger = logging.getLogger(__name__)

class TesseOCROptimized:
    """
    Optimized Tesseract OCR processor using tesserocr for maximum performance.
    
    Key improvements over pytesseract:
    - Model is loaded once and reused (memoization)
    - Direct C++ API access
    - No subprocess overhead
    - Thread-safe operations
    - Memory efficient processing
    """
    
    def __init__(self, language: str = 'deu+eng'):
        """
        Initialize optimized OCR processor.
        
        Args:
            language: Tesseract language code (e.g., 'deu+eng')
        """
        self.language = language
        self._lock = threading.Lock()
        self.processed_boxes = 0
        self.total_processing_time = 0.0
        
        # Set tessdata path for nix environment
        import os
        tessdata_path = "/nix/store/3xz7i2zscqhfp70fylqs04cn8y02frfs-tesseract-5.5.1/share/tessdata"
        os.environ['TESSDATA_PREFIX'] = tessdata_path
        
        try:
            # Initialize Tesseract API - this is the key performance improvement!
            self.api = tesserocr.PyTessBaseAPI(lang=language, path=tessdata_path)
            
            # Set optimal parameters for medical text recognition
            self.api.SetPageSegMode(tesserocr.PSM.SINGLE_WORD)  # PSM 8 for single words
            self.api.SetVariable('tessedit_char_whitelist', 
                               '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-üöäÜÖÄß.:/ ')
            
            # Optimize for speed and accuracy
            self.api.SetVariable('tessedit_enable_dict_correction', '1')
            self.api.SetVariable('tessedit_enable_bigram_correction', '1')
            
            logger.info("TesseOCR processor initialized with language: %s", language)
            
        except Exception as e:
            logger.error("Failed to initialize TesseOCR: %s", e)
            self.api = None
            raise
    
    def __del__(self):
        """Clean up Tesseract API."""
        if hasattr(self, 'api') and self.api:
            try:
                self.api.End()
            except Exception:
                pass
    
    def process_image_boxes(
        self, 
        image_path: Union[str, Image.Image], 
        boxes: List[Tuple[int, int, int, int]]
    ) -> Tuple[List[Tuple[str, Tuple[int, int, int, int]]], List[float]]:
        """
        Process multiple text boxes in an image using optimized Tesseract.
        
        Args:
            image_path: Path to image file or PIL Image object
            boxes: List of bounding boxes (startX, startY, endX, endY)
            
        Returns:
            Tuple of (extracted_text_with_boxes, confidences)
        """
        if not self.api:
            logger.error("TesseOCR API not initialized")
            return [], []
        
        # Load image
        if hasattr(image_path, "convert"):
            image = image_path.convert("RGB")
        else:
            image = Image.open(image_path).convert("RGB")
        
        extracted_text_with_boxes = []
        confidences = []
        
        start_time = time.time()
        logger.debug(f"Processing {len(boxes)} boxes with TesseOCR")
        
        with self._lock:  # Thread safety for the shared API
            for idx, box in enumerate(boxes):
                try:
                    (startX, startY, endX, endY) = box
                    
                    # Crop the image to the box
                    cropped_image = image.crop((startX, startY, endX, endY))
                    
                    # Set image in Tesseract API - much faster than subprocess!
                    self.api.SetImage(cropped_image)
                    
                    # Extract text directly from C++ API
                    ocr_result = self.api.GetUTF8Text().strip()
                    
                    # Get confidence score directly
                    confidence_score = self.api.MeanTextConf()
                    confidence_normalized = confidence_score if confidence_score > 0 else 0.0
                    
                    # Append results
                    extracted_text_with_boxes.append((ocr_result, box))
                    confidences.append(confidence_normalized)
                    
                    logger.debug(f"Processed box {idx + 1}/{len(boxes)}: "
                               f"'{ocr_result}' with confidence {confidence_normalized:.2f}")
                    
                except Exception as e:
                    logger.warning(f"Error processing box {idx + 1}/{len(boxes)}: {e}")
                    extracted_text_with_boxes.append(("", box))
                    confidences.append(0.0)
        
        # Update performance metrics
        processing_time = time.time() - start_time
        self.processed_boxes += len(boxes)
        self.total_processing_time += processing_time
        
        logger.info(f"TesseOCR processing complete: {len(boxes)} boxes in {processing_time:.3f}s "
                   f"({processing_time/len(boxes):.3f}s/box)")
        
        return extracted_text_with_boxes, confidences
    
    def get_performance_stats(self) -> dict:
        """Get performance statistics."""
        if self.processed_boxes == 0:
            return {'avg_processing_time': 0.0, 'total_boxes': 0}
        
        return {
            'avg_processing_time': self.total_processing_time / self.processed_boxes,
            'total_boxes': self.processed_boxes,
            'total_time': self.total_processing_time,
            'boxes_per_second': self.processed_boxes / self.total_processing_time if self.total_processing_time > 0 else 0
        }


# Global instance for reuse across multiple calls
_global_tesseocr_processor = None

def get_global_tesserocr_processor(language: str = 'deu+eng') -> TesseOCROptimized:
    """Get or create global TesseOCR processor for maximum efficiency."""
    global _global_tesseocr_processor
    
    if _global_tesseocr_processor is None:
        _global_tesseocr_processor = TesseOCROptimized(language)
    
    return _global_tesseocr_processor


def tesseract_on_boxes_fast(
    image_path: Union[str, Image.Image], 
    boxes: List[Tuple[int, int, int, int]],
    language: str = 'deu+eng'
) -> Tuple[List[Tuple[str, Tuple[int, int, int, int]]], List[float]]:
    """
    Drop-in replacement for tesseract_on_boxes with significant performance improvement.
    
    This function provides the same interface as the original pytesseract-based function
    but uses tesserocr for 10-50x faster processing.
    
    Args:
        image_path: Path to image file or PIL Image object
        boxes: List of bounding boxes (startX, startY, endX, endY)
        language: Tesseract language code
        
    Returns:
        Tuple of (extracted_text_with_boxes, confidences)
    """
    processor = get_global_tesserocr_processor(language)
    return processor.process_image_boxes(image_path, boxes)


def cleanup_global_processor():
    """Clean up global processor to free memory."""
    global _global_tesseocr_processor
    if _global_tesseocr_processor:
        del _global_tesseocr_processor
        _global_tesseocr_processor = None


# Performance comparison function for testing
def compare_ocr_performance(
    image_path: Union[str, Image.Image], 
    boxes: List[Tuple[int, int, int, int]],
    language: str = 'deu+eng'
) -> dict:
    """
    Compare performance between pytesseract and tesserocr implementations.
    
    Args:
        image_path: Path to image file or PIL Image object
        boxes: List of bounding boxes
        language: Tesseract language code
        
    Returns:
        Dictionary with performance comparison results
    """
    # Test TesseOCR
    start_time = time.time()
    tesserocr_results, tesserocr_confidences = tesseract_on_boxes_fast(image_path, boxes, language)
    tesserocr_time = time.time() - start_time
    
    # Test pytesseract (if available)
    pytesseract_time = None
    pytesseract_results = None
    pytesseract_confidences = None
    
    try:
        from lx_anonymizer.ocr import tesseract_on_boxes
        start_time = time.time()
        pytesseract_results, pytesseract_confidences = tesseract_on_boxes(image_path, boxes)
        pytesseract_time = time.time() - start_time
    except ImportError:
        logger.warning("pytesseract implementation not available for comparison")
    
    comparison = {
        'tesserocr': {
            'time': tesserocr_time,
            'results_count': len(tesserocr_results),
            'avg_confidence': sum(tesserocr_confidences) / len(tesserocr_confidences) if tesserocr_confidences else 0,
        }
    }
    
    if pytesseract_time is not None:
        comparison['pytesseract'] = {
            'time': pytesseract_time,
            'results_count': len(pytesseract_results) if pytesseract_results else 0,
            'avg_confidence': sum(pytesseract_confidences) / len(pytesseract_confidences) if pytesseract_confidences else 0,
        }
        
        comparison['speedup'] = pytesseract_time / tesserocr_time if tesserocr_time > 0 else float('inf')
        
        logger.info(f"Performance comparison: TesseOCR is {comparison['speedup']:.1f}x faster than pytesseract")
    
    return comparison

==> lx_anonymizer-0.8.3.tar.gz: lx_anonymizer-0.8.3/lx_anonymizer/ocr_tesserocr.py <==
"""
Optimized OCR module using Tesserocr for significantly improved performance.

This module replaces pytesseract usage with Tesserocr for:
- 10-50x faster text recognition in video processing
- Pre-loaded Tesseract model (memoization)
- Direct C++ API access without CLI overhead
- Better memory management for batch processing

Usage:
    from lx_anonymizer.ocr_tesserocr import tesseract_on_boxes_fast, TesseOCROptimized

    # Drop-in replacement for existing tesseract_on_boxes function
    results, confidences = tesseract_on_boxes_fast(image_path, boxes)

    # Or use the class directly for more control
    ocr_processor = TesseOCROptimized()
    results, confidences = ocr_processor.process_image_boxes(image_path, boxes)
"""

import logging
import threading
import time
from typing import List, Tuple, Union

import tesserocr
from PIL import Image

logger = logging.getLogger(__name__)


class TesseOCROptimized:
    """
    Optimized Tesseract OCR processor using tesserocr for maximum performance.

    Key improvements over pytesseract:
    - Model is loaded once and reused (memoization)
    - Direct C++ API access
    - No subprocess overhead
    - Thread-safe operations
    - Memory efficient processing
    """

    def __init__(self, language: str = "deu+eng"):
        """
        Initialize optimized OCR processor.

        Args:
            language: Tesseract language code (e.g., 'deu+eng')
        """
        self.language = language
        self._lock = threading.Lock()
        self.processed_boxes = 0
        self.total_processing_time = 0.0

        # Set tessdata path for nix environment
        import os

        tessdata_path = (
            "/nix/store/3xz7i2zscqhfp70fylqs04cn8y02frfs-tesseract-5.5.1/share/tessdata"
        )
        os.environ["TESSDATA_PREFIX"] = tessdata_path

        try:
            # Initialize Tesseract API - this is the key performance improvement!
            self.api = tesserocr.PyTessBaseAPI(lang=language, path=tessdata_path)

            # Set optimal parameters for medical text recognition
            self.api.SetPageSegMode(tesserocr.PSM.SINGLE_WORD)  # PSM 8 for single words
            self.api.SetVariable(
                "tessedit_char_whitelist",
                "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-üöäÜÖÄß.:/ ",
            )

            # Optimize for speed and accuracy
            self.api.SetVariable("tessedit_enable_dict_correction", "1")
            self.api.SetVariable("tessedit_enable_bigram_correction", "1")

            logger.info("TesseOCR processor initialized with language: %s", language)

        except Exception as e:
            logger.error("Failed to initialize TesseOCR: %s", e)
            self.api = None
            raise

    def __del__(self):
        """Clean up Tesseract API."""
        if hasattr(self, "api") and self.api:
            try:
                self.api.End()
            except Exception:
                pass

    def process_image_boxes(
        self,
        image_path: Union[str, Image.Image],
        boxes: List[Tuple[int, int, int, int]],
    ) -> Tuple[List[Tuple[str, Tuple[int, int, int, int]]], List[float]]:
        """
        Process multiple text boxes in an image using optimized Tesseract.

        Args:
            image_path: Path to image file or PIL Image object
            boxes: List of bounding boxes (startX, startY, endX, endY)

        Returns:
            Tuple of (extracted_text_with_boxes, confidences)
        """
        if not self.api:
            logger.error("TesseOCR API not initialized")
            return [], []

        # Load image
        if hasattr(image_path, "convert"):
            image = image_path.convert("RGB")
        else:
            image = Image.open(image_path).convert("RGB")

        extracted_text_with_boxes = []
        confidences = []

        start_time = time.time()
        logger.debug(f"Processing {len(boxes)} boxes with TesseOCR")

        with self._lock:  # Thread safety for the shared API
            for idx, box in enumerate(boxes):
                try:
                    (startX, startY, endX, endY) = box

                    # Crop the image to the box
                    cropped_image = image.crop((startX, startY, endX, endY))

                    # Set image in Tesseract API - much faster than subprocess!
                    self.api.SetImage(cropped_image)

                    # Extract text directly from C++ API
                    ocr_result = self.api.GetUTF8Text().strip()

                    # Get confidence score directly
                    confidence_score = self.api.MeanTextConf()
                    confidence_normalized = (
                        confidence_score if confidence_score > 0 else 0.0
                    )

                    # Append results
                    extracted_text_with_boxes.append((ocr_result, box))
                    confidences.append(confidence_normalized)

                    logger.debug(
                        f"Processed box {idx + 1}/{len(boxes)}: "
                        f"'{ocr_result}' with confidence {confidence_normalized:.2f}"
                    )

                except Exception as e:
                    logger.warning(f"Error processing box {idx + 1}/{len(boxes)}: {e}")
                    extracted_text_with_boxes.append(("", box))
                    confidences.append(0.0)

        # Update performance metrics
        processing_time = time.time() - start_time
        self.processed_boxes += len(boxes)
        self.total_processing_time += processing_time

        logger.info(
            f"TesseOCR processing complete: {len(boxes)} boxes in {processing_time:.3f}s "
            f"({processing_time / len(boxes):.3f}s/box)"
        )

        return extracted_text_with_boxes, confidences

    def get_performance_stats(self) -> dict:
        """Get performance statistics."""
        if self.processed_boxes == 0:
            return {"avg_processing_time": 0.0, "total_boxes": 0}

        return {
            "avg_processing_time": self.total_processing_time / self.processed_boxes,
            "total_boxes": self.processed_boxes,
            "total_time": self.total_processing_time,
            "boxes_per_second": self.processed_boxes / self.total_processing_time
            if self.total_processing_time > 0
            else 0,
        }


# Global instance for reuse across multiple calls
_global_tesseocr_processor = None


def get_global_tesserocr_processor(language: str = "deu+eng") -> TesseOCROptimized:
    """Get or create global TesseOCR processor for maximum efficiency."""
    global _global_tesseocr_processor

    if _global_tesseocr_processor is None:
        _global_tesseocr_processor = TesseOCROptimized(language)

    return _global_tesseocr_processor


def tesseract_on_boxes_fast(
    image_path: Union[str, Image.Image],
    boxes: List[Tuple[int, int, int, int]],
    language: str = "deu+eng",
) -> Tuple[List[Tuple[str, Tuple[int, int, int, int]]], List[float]]:
    """
    Drop-in replacement for tesseract_on_boxes with significant performance improvement.

    This function provides the same interface as the original pytesseract-based function
    but uses tesserocr for 10-50x faster processing.

    Args:
        image_path: Path to image file or PIL Image object
        boxes: List of bounding boxes (startX, startY, endX, endY)
        language: Tesseract language code

    Returns:
        Tuple of (extracted_text_with_boxes, confidences)
    """
    processor = get_global_tesserocr_processor(language)
    return processor.process_image_boxes(image_path, boxes)


def cleanup_global_processor():
    """Clean up global processor to free memory."""
    global _global_tesseocr_processor
    if _global_tesseocr_processor:
        del _global_tesseocr_processor
        _global_tesseocr_processor = None


# Performance comparison function for testing
def compare_ocr_performance(
    image_path: Union[str, Image.Image],
    boxes: List[Tuple[int, int, int, int]],
    language: str = "deu+eng",
) -> dict:
    """
    Compare performance between pytesseract and tesserocr implementations.

    Args:
        image_path: Path to image file or PIL Image object
        boxes: List of bounding boxes
        language: Tesseract language code

    Returns:
        Dictionary with performance comparison results
    """
    # Test TesseOCR
    start_time = time.time()
    tesserocr_results, tesserocr_confidences = tesseract_on_boxes_fast(
        image_path, boxes, language
    )
    tesserocr_time = time.time() - start_time

    # Test pytesseract (if available)
    pytesseract_time = None
    pytesseract_results = None
    pytesseract_confidences = None

    try:
        from lx_anonymizer.ocr import tesseract_on_boxes

        start_time = time.time()
        pytesseract_results, pytesseract_confidences = tesseract_on_boxes(
            image_path, boxes
        )
        pytesseract_time = time.time() - start_time
    except ImportError:
        logger.warning("pytesseract implementation not available for comparison")

    comparison = {
        "tesserocr": {
            "time": tesserocr_time,
            "results_count": len(tesserocr_results),
            "avg_confidence": sum(tesserocr_confidences) / len(tesserocr_confidences)
            if tesserocr_confidences
            else 0,
        }
    }

    if pytesseract_time is not None:
        comparison["pytesseract"] = {
            "time": pytesseract_time,
            "results_count": len(pytesseract_results) if pytesseract_results else 0,
            "avg_confidence": sum(pytesseract_confidences)
            / len(pytesseract_confidences)
            if pytesseract_confidences
            else 0,
        }

        comparison["speedup"] = (
            pytesseract_time / tesserocr_time if tesserocr_time > 0 else float("inf")
        )

        logger.info(
            f"Performance comparison: TesseOCR is {comparison['speedup']:.1f}x faster than pytesseract"
        )

    return comparison
    return comparison
